from gevent import monkey
monkey.patch_all(thread=False)
# from psycogreen.gevent import patch_psycopg
# patch_psycopg()

import os
from datetime import timedelta, datetime
# from flask_debugtoolbar import DebugToolbarExtension

from flask_cors import CORS
from flask_wtf import CSRFProtect
#from k2.k2rout import *
from flask import Flask, request, session, g, current_app
from flask_session import Session
from flask_login import current_user, LoginManager, logout_user
from flask_socketio import SocketIO, emit
#import eventlet
from ua_parser import user_agent_parser
import psutil
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate


from k2.k2logging import configure_logging
from k2.k2trans import K2Babel
from k2.k2cfg import K2
from k2.k2obj import K2Obj


# configure flask app
class K2Flask(Flask):
    def _check_setup_finished(self, f_name: str) -> None:
        pass

# from prometheus_flask_exporter import PrometheusMetrics


# configure flask app
app = K2Flask(__name__)


#app = Flask(__name__)
CORS(app)
app.config['SECRET_KEY'] = K2.secret_key
app.json.sort_keys = K2.json_sort_keys
#app.config.from_object(k2)
app.config['WTF_CSRF_CHECK_DEFAULT'] = False
csrf = CSRFProtect(app)
#eventlet.monkey_patch(socket=True, select=True) #для роботи в режимі eventlet
socketio = SocketIO(app, async_mode='gevent', cors_allowed_origins="*", manage_session=True, ping_timeout=180) #, async_mode='eventlet', cors_allowed_origins="*", policy_server=False)
app.config['SESSION_TYPE'] = 'filesystem'
app.config['SESSION_FILE_DIR'] = '../sessions'
Session(app)

# configure PYTHONPATH
pythonpath = os.environ.get('PYTHONPATH', '')
pythonpath += ':.'
os.environ['PYTHONPATH'] = pythonpath
# configure FLASK_APP
os.environ['FLASK_APP'] = 'app.py'


#db
app.config['SQLALCHEMY_DATABASE_URI'] = K2().init_db()
app.config['SQLALCHEMY_POOL_SIZE'] = K2.sglalchemy_pool_size
app.config['SQLALCHEMY_MAX_OVERFLOW'] = K2.sglalchemy_max_overflow
app.config['REMEMBER_COOKIE_DURATION'] = timedelta(days=7)
app.config['PERMANENT_SESSION_LIFETIME'] = timedelta(minutes=180)
app.jinja_env.trim_blocks = True
app.jinja_env.lstrip_blocks = True


db = SQLAlchemy()
db.init_app(app)
K2.db = db
with app.app_context():
    K2Obj.db = db.engines[None]


#logging
configure_logging()

#migrations
migrate = Migrate(app, db, autogenerate=True)  #directory='../migrations'

#translations
K2.load_babel_translation_directories()
app.config['BABEL_TRANSLATION_DIRECTORIES'] = K2.babel_translation_directories
if K2.current_language == '':
    app.config['BABEL_DEFAULT_LOCALE'] = K2.default_language
else:
    app.config['BABEL_DEFAULT_LOCALE'] = K2.current_language
app.config['BABEL_LANGUAGES'] = K2.languages
app.config['BABEL_DEFAULT_TIMEZONE'] = K2.timezone
#print(app.extensions)
babel = K2Babel(app)
babel.init_app(app, locale_selector=K2.get_locale)


#load components list
K2.component_list()


k2inst = K2()
@app.before_request
def before_request():
    if request.endpoint:
        if  'static' not in str(request.endpoint):
            g.start_time = datetime.now()
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                k2inst.check_logout_users(current_user.get_id())
                if current_user.is_authenticated and K2.db_login:
                    k2inst.db_user_engine(db)


@app.context_processor
def inject_static_variable():
    k2inst = K2()
    k2inst.search_menu_items()
    if K2.menu_category == []:
        k2inst.search_menu_items_category()
    if K2.class_dict == {}:
        k2inst.search_class_dict()
    server_software = request.environ.get('SERVER_SOFTWARE', '')

    if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                    LoginManager) and current_user.is_authenticated:
        error_dict = K2.load_logging_messages(request.url_rule)
        error_count = error_dict['error_count']
        warning_count = error_dict['warning_count']
        error_messages = error_dict['error_messages']
    else:
        error_count = 0
        warning_count = 0
        error_messages = 0
    
    context = {
        'menu': K2.menu,
        'menu_category': K2.menu_category,
        'template_name': k2inst.current_template(),
        'current_user_login': k2inst.current_user_login,
        'current_user_role': K2.get_current_user_role_name(),
        'current_user_permissions':  K2.get_user_permissions(),
        'locale' : K2.default_language,
        'request': request,
        'user_agent': user_agent_parser.Parse(request.headers.get('User-Agent')),
        'count_update': 0,
        'port': K2.port,
        'proj': K2.proj_config,
        'countreparts': k2inst.get_user_counterparts_name(),
        'storages': k2inst.get_user_stoages_name(),
        'project_name': k2inst.get_user_project_name(),
        'end_time': datetime.now(),
        'domain': K2.domain,
        'domain_comp': '',
        'online_users': k2inst.get_authorized_users(),
        'page_id': K2.namemenu(request.path),
        'error_count': error_count,
        'warning_count': warning_count,
        'error_messages': error_messages,
        'system_settings': K2.system_settings
        }
    if 'gunicorn' in server_software.lower():
        context['domain_comp'] = K2.domain
        # context['domain_comp'] = K2.domain + '/components' #для запуску з nginx
    return context
#from gevent.util import print_run_info, format_run_info


#turn on jinja cashe
if K2.system_settings and K2.system_settings['caching_enabled']:
    from jinja2 import Environment, FileSystemLoader, FileSystemBytecodeCache
    cache_dir = '../caсhe'
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    bytecode_cache = FileSystemBytecodeCache(directory=cache_dir)
    app.jinja_env.bytecode_cache = bytecode_cache


@app.after_request
# @by_path_counter
def after_request(response):
    try:
        if current_user.is_authenticated:
            k2inst.get_menu_url()
            all_routes = K2.menu_url
            if request.path in all_routes:
                #k2inst.add_authorized_users(current_user.get_id(), current_user.get_login())
                #for users in k2inst.get_authorized_users():
                    #k2inst.check_authorized_users(users['user_id'])
                try:
                    from components.k2site.k2site.views import K2Site
                    K2Site().user_login_save_log(current_user,
                                                 user_agent_parser.Parse(request.headers.get('User-Agent')),
                                                 request.headers.get('X-Forwarded-For', request.remote_addr),
                                                 'Open page',
                                                 request.path, psutil.cpu_percent(interval=1), psutil.virtual_memory().percent, psutil.disk_usage('/').percent )
                except:
                     pass
        else:
            if 'static' not in str(request.endpoint) or request.endpoint == 'None':
                from components.k2site.k2site.views import K2Site
                K2Site().user_login_save_log(current_user,
                                             user_agent_parser.Parse(request.headers.get('User-Agent')),
                                             request.headers.get('X-Forwarded-For', request.remote_addr),
                                             'Open page',
                                             request.path, psutil.cpu_percent(interval=1), psutil.virtual_memory().percent,
                                             psutil.disk_usage('/').percent)
    except:
        pass
    return response

#platform
K2.platform = K2.get_platform()


# flask routing
from routes import *


if __name__ == '__main__':
    # app.config['DEBUG_TB_PROFILER_ENABLED'] = True
    # app.debug = True
    # toolbar = DebugToolbarExtension(app)
    # app.run(use_reloader=False,   port=K2.port)
    socketio.run(app, port=int(K2.port))

