import yaml
import inspect
import pytz
import os
import sys
import platform
import uuid
from datetime import datetime, timedelta
import json

from flask import session, request, redirect, url_for,  send_from_directory
from flask_login import LoginManager, current_user, logout_user
from sqlalchemy.orm import joinedload, sessionmaker
from sqlalchemy import desc, create_engine
import logging

from k2.k2obj import K2Obj
from k2.k2secur import K2Secur
from k2.k2path import K2Path
from k2.k2datasync import K2Datasync
from k2.k2settings import K2Settings


class K2(K2Obj):
    name = 'k2cfg'
    version = '2.5.1.46'
    port = ''
    domain_name = ''
    domain = ''
    update_domain = ''
    components_install = []
    menu = []
    menu_url = []
    menu_category = []
    sglalchemy_pool_size = 15
    sglalchemy_max_overflow = 30
    safe_mode = False
    json_sort_keys = False
    babel_translation_directories = ''
    default_language = 'uk'
    current_language = 'uk'
    login_manager = LoginManager()
    timezone = pytz.timezone('Europe/Kiev')
    secret_key = 'JH&INH987gFDHdsagh&8dwbjdw8ckw'
    generate_migration = ''
    upgrade_migrations = ''
    platform = ''
    user_ip = ''
    user_agent = ''
    languages = {
        'en': 'English',
        'uk': 'Українська',
        'pl': 'Polski'
    }
    proj_config = yaml.safe_load(open("../proj.yml"))['proj']
    db = ''
    db_driver = ''
    class_dict = {}
    connected_clients = {}
    system_settings = {}
    adm_menu_items = [
        {
            "title": "t_update_system",
            "icon": {"icon": "mdi mdi-cloud-download"},
            "children": [
                {
                    "title": "t_install_new_components2",
                    "to": "/components-add"
                },
                {
                    "title": "t_installed_components2",
                    "to": "/components-list"
                },
                {
                    "title": "t_update_component",
                    "to": "/components-update"

                },
                {
                    "title": "t_lnSyncData",
                    "to": "/components-update-metadata"

                },
                {
                    "title": "t_lnSyncDataLicense",
                    "to": "/components-license-key"

                }

            ]
        }
    ]
    adm_menu_items_category = [
        {
            "title": 't_update_system',  # назва модуля
            "to": "/components-add",  # посилання на головну сторінку модуля
            "icon": "k2",  # назва іконки модуля
            "sort": 100
        }
    ]

    # шляхи до файлу з підключенням до БД
    path_db = f"../cfg/{proj_config}/db"
    path_db_user = f"../cfg/k2/db/users/"
    db_login = None
    static_files = []
    caching_enabled = None

    search_comp = [
        f"../cfg/{proj_config}/components",
        f"../usr/{proj_config}/components",
        f"components"
    ]
    app_root_path = ''
    template = 'app_stack'
    authorized_users = {}
    authorized_users_file = '../data/k2_cloud_erp/users/authorized_users.json'


    def __init__(self, *args, **kwargs):
        super().__init__()
        self.current_user_login = self.get_current_user()
        self.current_user_class = current_user
        self.secur = K2Secur()
        self.path = K2Path()
        self.data = K2Datasync()
        self.settings = K2Settings()

        #self.db = None

        #self.menu = self.search_menu_items()



    def init_db_uri(self):
        '''Пошук і ініціалізація файлу з підключення до бази даних '''
        #for path in self.path_db:
        try:
            db_file_path = self.path_db + '/db.yml'
            if os.path.exists(db_file_path) and os.path.getsize(db_file_path) > 0:
                with open(db_file_path, 'r') as f:
                    db_config = yaml.load(f, Loader=yaml.FullLoader)
                return db_config
            return None
        except Exception as e:
            logging.error(f"{e}")
            return None


    def init_db_uri_user(self):
        '''Пошук і ініціалізація файлу з підключення до бази даних для поточного користуача'''
        try:
            db_file_path = self.path_db_user + self.get_current_user().get_id() + '/db.yml'
            if os.path.exists(db_file_path) and os.path.getsize(db_file_path) > 0:
                with open(db_file_path, 'r') as f:
                    db_config = yaml.load(f, Loader=yaml.FullLoader)
                return db_config
            return None
        except Exception as e:
            logging.error(f"{e}")
            return None


    # creating a connections

    def init_db(self):
        '''Підключення db_uri для вибраної бази даних'''
        db_config = self.init_db_uri()
        if db_config is None:
            db_path = 'database.db'
            db_uri = f"sqlite:///{db_path}"
        else:
            #print(db_config)
            db_uri = f"{db_config['driver']}://{db_config['user']}:{db_config['password']}@{db_config['host']}:{db_config['port']}/{db_config['dbname']}"
        return db_uri


    def init_db_user(self):
        '''Підключення db_uri для користувача'''
        from app import app
        if (hasattr(app, 'login_manager')
                and isinstance(app.login_manager, LoginManager)
                and current_user
                and K2.db_login):
            self.create_db_file_config_user()
            db_config = self.init_db_uri_user()
            db_uri = f"{db_config['driver']}://{db_config['user']}:{db_config['password']}@{db_config['host']}:{db_config['port']}/{db_config['dbname']}"
            # engine = create_engine(db_uri)
            return db_uri


    def create_db_role(self, user_name, password):
        '''Створення користувача на рівні БД'''
        try:
            import psycopg2
            db_config = K2().init_db_uri()

            # Підключення до бази даних
            conn = psycopg2.connect(
                dbname=db_config['dbname'],
                user=db_config['user'],
                password=db_config['password'],
                host=db_config['host'],
                port=db_config['port']
            )

            # Створення курсора
            cur = conn.cursor()

            # Перевірка наявності користувача
            cur.execute("SELECT 1 FROM pg_roles WHERE rolname = %s", (user_name,))
            user_exists = cur.fetchone()
            if user_exists:
                return "Користувач вже існує"
            else:
                # Виконання SQL-запиту для створення користувача
                cur.execute(f"CREATE USER \"{user_name}\" WITH PASSWORD %s", (password,))
                cur.execute(f"GRANT ALL PRIVILEGES ON DATABASE \"{db_config['dbname']}\" TO \"{user_name}\"")
                cur.execute(f"ALTER USER \"{user_name}\" WITH SUPERUSER")

            # Підтвердження змін
            conn.commit()

            # Закриття курсора та з'єднання
            cur.close()
            conn.close()

            return "Користувач успішно створений"
        except Exception as e:
            logging.error(f"Помилка створення користувача {e}")
            return "Помилка створення користувача"


    def drop_db_role(self, user_name):
        '''Видалення користувача на рівні БД'''
        try:
            import psycopg2
            db_config = K2().init_db_uri()

            # Підключення до бази даних
            conn = psycopg2.connect(
                dbname=db_config['dbname'],
                user=db_config['user'],
                password=db_config['password'],
                host=db_config['host'],
                port=db_config['port']
            )

            # Створення курсора
            cur = conn.cursor()

            # Перевірка наявності користувача
            cur.execute("SELECT 1 FROM pg_roles WHERE rolname = %s", (user_name,))
            user_exists = cur.fetchone()
            if not user_exists:
                return "Користувач не існує"
            else:
                # Виконання SQL-запиту для видалення користувача
                cur.execute(f"REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA public FROM \"{user_name}\"")
                cur.execute(f"DROP OWNED BY \"{user_name}\"")
                cur.execute(f"DROP USER \"{user_name}\"")

            # Підтвердження змін
            conn.commit()

            # Закриття курсора та з'єднання
            cur.close()
            conn.close()
            return "Користувач успішно видалений"
        except Exception as e:
            logging.error(f"Помилка видалення користувача {e}")
            return "Помилка видалення користувача"


    def create_db_file_config_user(self):
        '''Створення файлу з параметрами підключенням до БД для користувача'''
        db_file_path = self.path_db_user + self.get_current_user().get_id() + '/db.yml'
        db_config = K2().init_db_uri()

        from app import app
        import hashlib
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                login = user.login
                password = user.password
        password = hashlib.md5(password.encode()).hexdigest()
        # Перевірка наявності шляху та створення відсутніх каталогів
        db_dir = os.path.dirname(db_file_path)
        if not os.path.exists(db_dir):
            os.makedirs(db_dir)
        # Перевірка наявності файлу та його створення, якщо він не існує
        if not os.path.exists(db_file_path):
            with open(db_file_path, 'w') as f:
                yaml.safe_dump({}, f)  # Створення пустого файлу

        # Зчитування поточного вмісту файлу
        with open(db_file_path, 'r') as f:
            existing_data = yaml.safe_load(f) or {}

        # Оновлення поточного вмісту файлу з даними конфігурації
        existing_data.update({
            'dbname': db_config['dbname'],
            'driver': db_config['driver'],
            'user': login,
            'password': password,
            'host': db_config['host'],
            'port': db_config['port']
        })
        # # Створення словника з даними конфігурації
        # data = {
        #     'dbname': db_config['dbname'],
        #     'user': login,
        #     'password': 'k2adm123456',
        #     'host': db_config['host'],
        #     'port': db_config['port']
        # }

        # Запис оновлених даних у файл YAML
        with open(db_file_path, 'w') as f:
            yaml.safe_dump(existing_data, f)
        self.db.session.close()


    def db_user_engine(self, db):
        '''Зміна параметрів підключення для поточного користувача'''

        if current_user.get_login() not in db.engines:
            db.engine.dispose()
            db.engines[current_user.get_login()] = create_engine(self.init_db_user())
            #db.engines[None] = create_engine(self.init_db_user())  #
        K2Obj.db = db.engines[current_user.get_login()]
        #return db.engines[self.current_user_class.get_login()]


    def init_lm(self, app):
        '''Ініціалізація логін-менеджера'''
        # Initialize LoginManager
        self.login_manager.init_app(app)


    def get_current_user(self):
        '''Select current_user
        K2().get_current_user()'''

        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager, LoginManager):
            return current_user
        else:
            return None

    @classmethod
    def create_system_settings(cls):
        # Дані для збереження налаштувань в JSON-файл
        data_settings = {
            "site_name": "K2 CloudERP",
            "logo_path": '',
            "caching_enabled": False
        }
        # Шлях до файлу для збереження налаштувань
        file_system_settings = '../cfg/k2/settings/system_settings.json'
        # Перевірка існування файлу
        if not os.path.exists(file_system_settings):
            # Створення директорій, якщо вони не існують
            os.makedirs(os.path.dirname(file_system_settings), exist_ok=True)

            # Створення та запис JSON-файлу
            with open(file_system_settings, 'w') as file:
                json.dump(data_settings, file, indent=4, ensure_ascii=False)

    def add_authorized_users(self, user_id, login):
        '''Додати користувача до списку авторизованих'''

        # Перевірити наявність каталогу
        directory = os.path.dirname(self.authorized_users_file)
        if not os.path.exists(directory):
            os.makedirs(directory)

        # Якщо файл не існує, створити його та ініціалізувати порожній список
        if not os.path.exists(self.authorized_users_file):
            with open(self.authorized_users_file, 'w') as f:
                json.dump([], f)

        # Зчитати існуючі дані з файлу
        with open(self.authorized_users_file, 'r') as f:
            data = json.load(f)

        # Перевірити, чи вже є користувач з таким user_id
        user_exists = any(user['user_id'] == user_id for user in data)

        # Якщо користувача ще немає, додати його
        if not user_exists:
            data.append({'user_id': user_id, 'login': login})
            with open(self.authorized_users_file, 'w') as f:
                json.dump(data, f, indent=4)


    def dell_authorized_users(self, user_id):
        '''Видалити користувача зі списку авторизованих'''
        if os.path.exists(self.authorized_users_file):
            with open(self.authorized_users_file, 'r') as f:
                data = json.load(f)
            # Видалити користувача зі списку
            data = [user for user in data if user['user_id'] != user_id]
            # Записати оновлені дані у файл
            with open(self.authorized_users_file, 'w') as f:
                json.dump(data, f)

    def kill_user_sessions(self, target_username):
        '''Завершення  сесії підключення користувача на рівні БД'''
        try:
            if K2.db_login:
                if K2Obj.db.name == 'postgresql':
                    with K2Obj.db.connect() as conn:  # Використання контекстного менеджера
                        try:
                            with conn.connection.cursor() as cur:  # Використання внутрішнього з'єднання для курсора
                                # Отримання активних сесій
                                cur.execute("""
                                    SELECT pid, usename, client_addr, application_name 
                                    FROM pg_stat_activity
                                    WHERE usename = %s;
                                """, (target_username,))

                                sessions = cur.fetchall()
                                print("Active sessions for user '{}':".format(target_username))
                                for session in sessions:
                                    print(session)

                                # Завершення всіх сесій певного користувача
                                cur.execute("""
                                    SELECT pg_terminate_backend(pid)
                                    FROM pg_stat_activity
                                    WHERE usename = %s;
                                """, (target_username,))

                                terminated_sessions = cur.fetchall()
                                print("Terminated sessions for user '{}':".format(target_username))
                                for terminated in terminated_sessions:
                                    print(terminated)

                            # Коміт змін
                            conn.commit()
                        except Exception as e:
                            print("Error:", e)
                            conn.rollback()
        except Exception as e:
            logging.error(f'Error: {e}')


    # def is_logout_authorized_users(self, user_id):
    #     '''Додавання мітки is_logout до списку авторизованих'''
    #     with open(self.authorized_users_file, 'r') as f:
    #             authorized_users = json.load(f)
    #     for user in authorized_users:
    #         if user['user_id'] == user_id:
    #             break
    #
    #     with open(self.authorized_users_file, 'w') as f:
    #         json.dump(authorized_users, f)

    def get_authorized_users(self):
        '''Отримати список авторизованих користувачів з файлу'''
        if os.path.exists(self.authorized_users_file):
            with open(self.authorized_users_file, 'r') as f:
                data = json.load(f)
            return data
        else:
            return []

        with open(self.authorized_users_file, 'w') as f:
            json.dump(authorized_users, f)


    def check_logout_users(self, user_id):
        try:
            from app import app
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                from components.k2site.k2site.models import K2users
                user_id = current_user.get_id()

                user_online_list_f = self.get_authorized_users()
                user_online_list = [user['user_id'] for user in user_online_list_f]
                user = K2users.query.filter_by(user_id=user_id).first()
                if user.locale is not None:
                    K2.current_language = user.locale
                else:
                    K2.current_language = K2.default_language
                if user_id not in user_online_list and user.is_logout_by_admin == 1:
                    logout_user()
                else:
                    self.add_authorized_users(current_user.get_id(), current_user.get_login())
                self.db.session.close()
        except Exception as e:
            logging.info(e)


    def check_authorized_users(self, user_id):
        '''Перевірити чи користувач активний протягом останніх 5 хв'''
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            try:
                from components.k2site.k2site.models import K2log_auth
                last_activity  =  K2log_auth.query.filter_by(user_id=user_id).order_by(desc(K2log_auth.date_auth)).first()
                if last_activity:
                    last_activity_date = last_activity.date_auth
                    if last_activity_date:
                        # Calculate the time difference between the last activity date and the current time
                        time_difference = datetime.now() - last_activity_date
                       # Check if the time difference is greater than 5 minutes
                       #  if time_difference > timedelta(minutes=0):
                       #      self.dell_authorized_users(user_id)
            except Exception as e:
                logging.error(f'Error check_authorized_users {e}')
            finally:
                self.db.session.close()


    # def get_macadress_for_ip(self, ip):
    #     'Returns a list of MACs for interfaces that have given IP, returns None if not found'
    #     import netifaces as nif
    #     for i in nif.interfaces():
    #         addrs = nif.ifaddresses(i)
    #         try:
    #             if_mac = addrs[nif.AF_LINK][0]['addr']
    #             if_ip = addrs[nif.AF_INET][0]['addr']
    #         except Exception as e:
    #             if_mac = if_ip = None
    #         if if_ip == ip:
    #             return if_mac
    #     return None

    # load database settings from yml file

    def get_user_project_id(self):
        '''Пошук id поточного проекту користувача
            K2().get_user_project_id()

        '''
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                current_proj_id = user.projid
            else:
                current_proj_id = None
        else:
            current_proj_id = None
        self.db.session.close()
        return current_proj_id


    def get_user_project_name(self):
        '''Пошук назви поточного проекту користувача
           K2().get_user_project_name()
        '''
        try:
            if self.get_user_project_id() :
                from components.k2site.k2site.models import K2Proj
                projid = self.get_user_project_id()
                project = K2Proj.query.get(projid)
                if project:
                    projname = project.projname
                else:
                    projname = None
            else:
                projname = None
            return projname
        except Exception as e:
            logging.error(f"{e}")
            return None
        finally:
            self.db.session.close()


    def get_user_counterparts_id(self):
        '''Пошук id поточного контрагента
           K2().get_user_counterparts_id()
           '''
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                current_counterpart_id = user.counterpart_id
            else:
                current_counterpart_id = None
        else:
            current_counterpart_id = None
        self.db.session.close()
        return current_counterpart_id


    def get_user_counterparts_name(self):
        '''Пошук назви поточного контрагента
           K2().get_user_counterparts_name()
        '''
        try:
            if self.get_user_counterparts_id() :
                from components.k2handbook.k2handbook.models import k2counterparts
                counterparts_id = self.get_user_counterparts_id()
                counterparts = k2counterparts.query.get(counterparts_id)
                if counterparts:
                    counterparts_name = counterparts.counterpart_name
                else:
                    counterparts_name = '-'
            else:
                counterparts_name = '-'
            return counterparts_name
        except Exception as e:
            logging.error(f"{e}")
            counterparts_name = '-'
            return counterparts_name
        finally:
            self.db.session.close()


    def get_user_storage_id(self):
        '''Пошук id поточного складу
            K2().get_user_storage_id()
        '''
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                current_storage_id = user.storage_id
            else:
                current_storage_id = None
        else:
            current_storage_id = None
        self.db.session.close()
        return current_storage_id


    def get_user_stoages_name(self):
        '''Пошук назви поточного складу
            K2().get_user_stoages_name()
        '''
        try:
            if self.get_user_storage_id() :
                from components.k2handbook.k2handbook.models import k2storage
                storage_id = self.get_user_storage_id()
                storages = k2storage.query.get(storage_id)
                if storages:
                    storages_name = storages.storage_name
                else:
                    storages_name = '-'
            else:
                storages_name = '-'
            return storages_name
        except Exception as e:
            logging.error(f"{e}")
            storages_name = '-'
            return storages_name
        finally:
            self.db.session.close()

    def get_project_setting(self):
        '''Форма зміни проекту'''
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            try:
                from components.k2site.k2site.views import K2Site
                grid_settings = K2Site.show_projects()
                return grid_settings
            except Exception as e:
                logging.error(f"{e}")
                grid_settings = ''
                return grid_settings

    def get_all_subclasses(cls):
        subclasses = K2Obj
        for subclass in list(subclasses):
            subclasses.extend(cls.get_all_subclasses(subclass))
        return subclasses

    def search_class_dict(self, parent_class=None):
        '''Рекурсивний пошук властивостей нащадків базового класу'''

        if parent_class is None:
            parent_class = K2Obj

        for subclass in parent_class.__subclasses__():
            class_name = subclass.__name__.lower()
            K2.class_dict[class_name] = {}
            # Додаємо властивості поточного підкласу до словника
            # if hasattr(subclass, 'adm_menu_items'):
            #     K2.class_dict[class_name]['adm_menu_items'] = subclass.adm_menu_items
            # if hasattr(subclass, 'adm_menu_items_category'):
            #     K2.class_dict[class_name]['adm_menu_items_category'] = subclass.adm_menu_items_category
            if hasattr(subclass, '_path_class'):
                K2.class_dict[class_name]['path'] = subclass._path_class
            if hasattr(subclass, 'export_data'):
                K2.class_dict[class_name]['export_data'] = subclass.export_data
            if hasattr(subclass, '_name_file_class'):
                K2.class_dict[class_name]['name_file'] = subclass._name_file_class

            # Рекурсивно шукаємо властивості нащадків поточного підкласу
            self.search_class_dict(subclass)


    def url_map(self):
        from flask import current_app
        all_routes = [
            rule.rule
            for rule in current_app.url_map.iter_rules()
        ]

        return all_routes

    def get_menu_url(self):
        '''Повертає список url актиивних пунктів меню'''
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from .k2upd import K2Menu
            response = K2.menu
            K2.menu_url = K2Menu().get_admin_menu_url(response)


    def search_menu_items(self):
        '''Search menu items'''
        try:
            from app import app
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                from .k2upd import K2Menu
                menu = K2Menu()
                K2.menu = menu.get_admin_menu()
                #K2.menu_category = menu.get_admin_menu_category()  # requests.get(url=f"{K2.domain_loc}/api/main-menu-category/", headers=request.headers)

                #K2.menu_url = K2Menu().get_admin_menu_url(response)
                #url_map all routes
                # K2.menu_url = [
                #     rule
                #     for rule in current_app.url_map.iter_rules()
                # ]
            else:
                K2.menu = []
                #K2.menu_category = []

        except Exception as e:
            logging.error(f"search_menu_items {e}")


    def search_menu_items_category(self):
        '''Search menu items category'''
        try:
            from app import app
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                from k2.k2upd import K2Menu
                response = K2Menu().get_admin_menu_category()  # requests.get(url=f"{K2.domain_loc}/api/main-menu-category/", headers=request.headers)
                # if response.status_code == 200:
                K2.menu_category = response
                # else:
                #    self.menu_category = []
            else:
                K2.menu_category = []
        except Exception as e:
            logging.error(f"{e}")


    def current_template(self):
        '''Search current template'''
        current_template_bd = None
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                current_template_bd = user.user_theme_pyt
        if current_template_bd is not None:
            return current_template_bd
        else:
            return self.template
        self.db.session.close()

    # def get_user_agent(self, current_request):
    #     pass

    def search_static_files(self):
        components_names = K2.search_comp_names()
        data = [
            subclass.adm_menu_items_category[0]
            for subclass in K2Obj.__subclasses__()
            if hasattr(subclass, 'adm_menu_items_category') and subclass.__name__.lower() in components_names
        ]

    @classmethod
    def namemenu(cls, url):
        '''
        Add identificators to url page
        '''
        namemenu_itm = 'k2itm' + url.replace('/', '-')
        return namemenu_itm

    @classmethod
    def get_user_role(cls, user_id):
        '''Отрмання roleid користувача '''
        try:
            from app import app
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                from components.k2site.k2site.models import K2users
                from .k2admmenu import K2admin_Menus_Prava, K2admin_menus

            # Отримати roleid
                user_role = K2users.query.filter_by(user_id=user_id).first().roleid
                return user_role
        except Exception as ex:
             logging.error(f"Error in get_user_role {ex}")
             user_role = None
             return user_role
        finally:
            cls.db.session.close()


    @classmethod
    def get_current_user_role_name(cls):
        '''Отрмання rolename поточного користувача '''
        try:
            from app import app
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                from components.k2site.k2site.models import K2roles
            user_id = current_user.get_id()
            roleid = cls.get_user_role(user_id)
            # Отримати rolename
            user_role = K2roles.query.filter_by(roleid=roleid).first().rolename
            return user_role
        except Exception as ex:
             print(ex)
             return {'error': 'Помилка при виконанні запиту до бази даних'}
        finally:
            cls.db.session.close()

    @classmethod
    def get_user_permissions(cls):
        """Retrieve user permissions based on the URL.

    This method fetches permissions for the current user based on the URL they are accessing.

    Returns:
        dict: A dictionary containing the user permissions, including 'roleid' and various permission flags
            such as 'r' (read), 'w' (write), 'i' (insert), 'd' (delete), 'c' (create), 'exp' (export), 'imp' (import), 'del_' (undelete), 'settable' (set table), 'cutpast' (cut and paste), 'enable' (enable), and 'active' (active).
            If the user is a superadmin, a default set of permissions is returned.
            If no permissions are found for the user, an empty dictionary is returned.

    Raises:
        Exception: If an error occurs during database query execution.
    Usage:
        current_user_permissions = K2().get_user_permissions()
    """

        try:
            from app import app
            if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                            LoginManager) and current_user.is_authenticated:
                from components.k2site.k2site.models import K2users
                from .k2admmenu import K2admin_Menus_Prava, K2admin_menus

                # Отримати roleid
                user_id = current_user.get_id()
                # Session = sessionmaker(bind=cls.db.engine)
                # session = Session()
                if K2Secur().is_superadmin_cloud(current_user):
                    permissions_list = {'roleid': '-1000', 'r': 1, 'w': 1, 'i': 1, 'd': 1, 'c': 1, 'exp': 1, 'imp': 1,
                                        'del_': 1, 'settable': 1, 'cutpast': 1, 'enable': 1, 'active': 1}
                    return permissions_list
                else:
                    user_role = cls.get_user_role(user_id)
                    namemenu = K2.namemenu(request.path)
                    try:
                        # Отримати права за вказаним roleid та namemenu
                        user_permissions = (
                            K2admin_Menus_Prava.query
                            .join(K2admin_menus, K2admin_Menus_Prava.menuid == K2admin_menus.menuid)
                            .filter(K2admin_Menus_Prava.roleid == user_role, K2admin_menus.namemenu == namemenu)
                            .options(joinedload(K2admin_Menus_Prava.menu))
                            .all()
                        )
                        # Перетворити результат в словник
                        if user_permissions:
                            permissions_list = []
                            for permission in user_permissions:
                                permissions_dict = {
                                    'roleid': permission.roleid,
                                    'r': permission.r,
                                    'w': permission.w,
                                    'i': permission.i,
                                    'd': permission.d,
                                    'c': permission.c,
                                    'exp': permission.exp,
                                    'imp': permission.imp,
                                    'del_': permission.del_,
                                    'settable': permission.settable,
                                    'cutpast': permission.cutpast,
                                    'enable': permission.enable,
                                    'active': permission.active
                                    # Додайте інші поля за необхідності
                                }
                                permissions_list.append(permissions_dict)
                                return permissions_list[0]
                        else:
                            return {}
                    except Exception as e:
                        logging.error(f'get_user_permissions {e}')
                        return {}

        except Exception as e:
            logging.error(f'get_user_permissions {e}')


    @classmethod
    def search_yml(cls, name_yml):
        """Search for a yml file in the caller directories of subclasses.

        This method searches for a yml file with a specified name in the caller directories of all subclasses of K2Obj.
        The search is performed in the 'yml' subdirectory of each caller directory.

        Args:
            name_yml: The name of the yml file to search for (without the '.yml' extension).

        Returns:
            The caller directory where the yml file was found, or None if the file was not found.

        Usage:
            To search for a yml file named 'users.yml', call this method as follows:
                K2.search_yml('users')
        """
        search_data_folder=True
        for subclass in K2Obj.__subclasses__():
            caller_path = inspect.getfile(subclass)
            caller_dir = os.path.dirname(caller_path)

            # Search for the yml file in the 'yml' subdirectory
            yml_dir = os.path.join(caller_dir, 'yml')
            yml_file = os.path.join(yml_dir, f'{name_yml}.yml')
            if os.path.exists(yml_file):
                search_data_folder=False
                # Return the caller directory if the yml file is found
                return caller_dir
        if search_data_folder:
            yml_path=f'yml/{name_yml}.yml'
            caller_dir =K2().path.find_data_file(yml_path)
            if os.path.exists(caller_dir):
                return caller_dir.replace(yml_path,'')

        # Return None if the yml file is not found
        return None

    @classmethod
    def get_path_to_root(cls, caller_file):
        '''
        Відносний шлях до рута.
        Виклик  K2.get_path_to_root(__file__).
        __file__ - стандартна змінна в Python, яка містить абсолютний шлях до поточного файлу, в якому вона використовується.
        '''
        curr_dir = os.path.dirname(__file__)  # Каталог поточного файлу
        root_dir = os.path.dirname(curr_dir)  # Каталог батьківського каталогу
        relative_path = os.path.relpath(caller_file, root_dir)
        return relative_path

    @classmethod
    def get_path_abs(cls, caller_file):
        '''
        Абсолютний шлях до файлу, в якому виконується код.
        Виклик K2.get_path_abs(__file__)
        __file__ -стандартна змінна в Python, яка містить абсолютний шлях до поточного файлу, в якому вона використовується. '''
        return os.path.abspath(caller_file)


    @classmethod
    def generate_id(cls):
        '''Генерація ID'''
        generated_id = uuid.uuid4().hex[:32]
        return ''.join([generated_id[i:i + 4] for i in range(0, len(generated_id), 4)])


    @classmethod
    def load_babel_translation_directories(cls):
        '''Завантаження перекладів'''
        with open('languages/babel_translation_directories.yml', 'r') as f:
            data = yaml.safe_load(f)
            # print(data['babel_translation_directories'])
        cls.babel_translation_directories = data['babel_translation_directories']


    @classmethod
    def search_babel_translation_directories(cls):
        '''Пошук файлів з перекладами'''
        # response = requests.get(f'{cls.domain}api/languages')
        from k2.k2trans import K2Lang
        response = K2Lang.find_language()
        if response:
            cls.babel_translation_directories = response
            data = {'babel_translation_directories': response}
        else:
            cls.babel_translation_directories = 'languages'
            data = {'babel_translation_directories': 'languages'}

        with open('languages/babel_translation_directories.yml', 'w') as f:
            yaml.dump(data, f)


    @classmethod
    def get_locale(cls):
        '''Визначення поточної мови'''
        if 'lang' in session:
            # Отримати обрану мову з сеансу користувача
            return session['lang']
        elif cls.current_language != '':
            return cls.current_language
        else:
            # Повернути мову за замовчуванням з налаштувань додатку
            return cls.default_language  # app.config['DEFAULT_LANGUAGE']


    @classmethod
    def get_locale_id(cls):
        '''Визначення id поточної мови'''
        lang = cls.get_locale()
        try:
            Session = sessionmaker(bind=K2Obj.db)
            session = Session()
            from components.k2handbook.k2handbook.models import k2language

            # language = k2language.query.filter_by(language_symbol=lang).first()
            language = session.query(k2language).filter_by(language_symbol=lang).first()
            if language:
                lang_id = language.language_id
            else:
                lang_id = None
        except Exception as e:
            logging.error(f"Помилка при виконанні запиту до бази даних. {e}")
            lang_id = None
        finally:
            session.close()
        return lang_id


    @classmethod
    def get_active_lang_list(cls):
        '''Визначення id поточної мови'''
        try:
            Session = sessionmaker(bind=K2Obj.db)
            session = Session()
            from components.k2handbook.k2handbook.models import k2language
            lang_list = []
            # language = k2language.query.filter_by(language_symbol=lang).first()
            language = session.query(k2language).filter_by(language_active=1).all()
            if language:
                for lang in language:
                    lang_list.append(lang.language_symbol)
        except Exception as e:
            logging.error(f"Помилка при виконанні запиту до бази даних. {e}")
            lang_id = None
        finally:
           session.close()

        return lang_list

    @classmethod
    def create_system_settings(cls):
        # Дані для збереження налаштувань в JSON-файл
        data_settings = {
            "title": "K2 Cloud ERP",
            "description": "K2 Cloud ERP: WMS, CRM, Document Management, CMS, Education System, VDoc",
            "logo_path": "",
            "caching_enabled": False,
            "db_login": False,
            "default_language": "uk",
            "action_log": False
        }
        # Шлях до файлу для збереження налаштувань
        file_system_settings = '../cfg/k2/settings/system_settings.json'
        # Перевірка існування файлу
        if not os.path.exists(file_system_settings):
            # Створення директорій, якщо вони не існують
            os.makedirs(os.path.dirname(file_system_settings), exist_ok=True)

            # Створення та запис JSON-файлу
            with open(file_system_settings, 'w') as file:
                json.dump(data_settings, file, indent=4, ensure_ascii=False)

    @classmethod
    def get_platform(cls):
        '''Визначення операційної системи'''
        operating_system = platform.system()
        cls.platform = operating_system
        return cls.platform

    path_objs = [
        f"../cfg/{proj_config}/k2/{name.lower()}.py",
        f"cfg/{name.lower()}.py",
        f"usr/cfg/{name.lower()}.py"
    ]

    @classmethod
    def ins_search_comp(self):
        '''Вивід списку компонент з файлу components.yml у всіх шляхах пошуку '''
        result_dict = {'components': []}
        for components_list_yml in self.search_comp:
            if os.path.exists(components_list_yml):
                with open(f"{components_list_yml}/components.yml", 'r') as file:
                    components_data_itm = yaml.safe_load(file)

                    for component in components_data_itm['components']:
                        name = component['name']
                        existing_component = next(
                            (c for c in result_dict['components'] if c['name'] == name),
                            None
                        )
                        if existing_component is None:
                            result_dict['components'].append(component)
        return result_dict

    @classmethod
    def search_comp_names(self):
        '''Вивід списку ідентифікаторів компонент
            у всіх шляхах пошуку

            Rerurn: list of components identificators
        '''
        components_data = self.ins_search_comp()
        if components_data is None or 'components' not in components_data:
            components_d = []
        else:
            components_d = components_data['components']
        # Вибираємо лише компоненти, у яких значення 'installed' дорівнює True
        components = [component['name'] for component in components_d if component.get('installed', True)]

        return components

    @classmethod
    def dump_stacks(cls):
        """
        Request information about the running threads of the current process.

        :return: A tuple containing:
                 - A sequence of text lines detailing the stacks of running
                   threads and greenlets. (One greenlet will duplicate one thread,
                   the current thread and greenlet.)
                 - The number of greenlets.
        """
        dump = []
        greenlet_count = 0  # Лічильник зелених потоків

        import gc
        import threading
        import sys
        import traceback
        try:
            from greenlet import greenlet
        except ImportError:
            greenlet = None

        threads = {th.ident: th.name for th in threading.enumerate()}

        for thread, frame in sys._current_frames().items():
            thread_name = threads.get(thread)
            if thread_name:
                dump.append('Thread: %s\n' % thread_name)
                dump.append(''.join(traceback.format_stack(frame)))
                dump.append('\n')

        if greenlet:  # Перевіряємо, що модуль greenlet доступний
            for ob in gc.get_objects():
                if isinstance(ob, greenlet) and ob:
                    greenlet_count += 1
                    #dump.append('Greenlet %s\n' % ob)
                    #dump.append(''.join(traceback.format_stack(ob.gr_frame)))
                    #dump.append('\n')

        return greenlet_count  # Повертаємо інформацію про стеки та кількість зелених потоків

    @classmethod
    def compare_versions(cls,version1, version2):
        # Розділяємо версії на числа
        parts1 = list(map(int, version1.split('.')))
        parts2 = list(map(int, version2.split('.')))

        # Порівнюємо кожну частину версії
        for i in range(min(len(parts1), len(parts2))):
            if parts1[i] < parts2[i]:
                return -1  # Версія 1 менше версії 2
            elif parts1[i] > parts2[i]:
                return 1  # Версія 1 більше версії 2

        # Якщо всі частини досягли рівняння, то версії рівні
        if len(parts1) < len(parts2):
            return -1  # Версія 1 коротше версії 2
        elif len(parts1) > len(parts2):
            return 1  # Версія 1 довше версії 2

        return 0  # Версії ідентичні

    log_error = 'ERROR'
    log_warning = 'WARNING'
    log_success = 'SUCCESS'

    @classmethod
    def log_entry(cls, error_id, status, message, name):
        log_entry = {
            'error_id': error_id,
            'date': datetime.now().strftime("[%Y-%m-%d %H:%M:%S]"),
            'status': status,
            'page': name,
            'message': message,
            'short_message': message[:100] + "..." if len(message) > 100 else message + "..."
        }
        return log_entry

    @classmethod
    def clear_logging_messages(cls):
        """Clear all logging message by user"""
        logging_dir = os.path.dirname(K2().path.data_path('conf/error_log/', True))
        filename = f"error.json"
        logging_file_dir = os.path.join(logging_dir, filename)
        logs = []
        with open(logging_file_dir, 'w') as file:
            json.dump(logs, file, indent=4)


    @classmethod
    def save_logging_message(cls, error_id, status, message, name):
        """
        Save a logging message to the stack.

        Parameters:
        status (str): The status of the message (e.g., cls.log_error -  'ERROR', cls.log_warning - 'WARNING', cls.log_success -  'SUCCESS').
        message (str): The content of the message to be saved.
        name (str): Name of object where the message originated.

        """

        logging_dir = os.path.dirname(K2().path.data_path('conf/error_log/', True))
        if not os.path.exists(logging_dir):
            os.makedirs(logging_dir)

        user_id = current_user.get_id()
        filename = f"error.json"
        logging_file_dir = os.path.join(logging_dir, filename)

        log_entry = cls.log_entry(error_id, status, message, name)

        # Load existing logs
        if os.path.exists(logging_file_dir):
            with open(logging_file_dir, 'r') as file:
                logs = json.load(file)
        else:
            logs = []

        # Append new log entry
        logs.append(log_entry)

        # Keep only the last 10 entries
        #logs = logs[-10:]
        # Save updated logs
        with open(logging_file_dir, 'w') as file:
            json.dump(logs, file, indent=4)
        from app import socketio
        error_dict = cls.load_logging_messages(name)
        #
        pageid= K2.namemenu(name)
        if current_user.get_id() in K2.connected_clients:
            sid = K2.connected_clients[current_user.get_id()]
            socketio.emit(f'update_errors_count_result_{pageid}', {'error_count': error_dict['error_count'], 'warning_count': error_dict['warning_count']}, room=sid)
        else:
            socketio.emit(f'update_errors_count_result_{pageid}',  {'error_count': error_dict['error_count'], 'warning_count': error_dict['warning_count']}, room=current_user.get_id())



    @classmethod
    def load_logging_messages(cls, page_url: str):
        """
        Load logging messages for the current page from the error log file.

        Parameters:
        name (str): Name of the object (page) to load messages for.

        Returns:
        dict: A dictionary containing formatted log messages, the count of 'error' messages, and the count of 'warning' messages.
        """
        logging_dir = os.path.dirname(K2().path.data_path('conf/error_log/', True))
        filename = f"error.json"
        logging_file_dir = os.path.join(logging_dir, filename)

        if not os.path.exists(logging_file_dir):
            return {'error_messages': '', 'error_count': 0, 'warning_count': 0}

        # Load existing logs
        with open(logging_file_dir, 'r') as file:
            logs = json.load(file)

        # Filter logs for the current page
        filtered_logs = [log for log in logs if log['page'] == str(page_url)]

        # Format messages and count errors and warnings
        formatted_messages = []
        error_count = 0
        warning_count = 0
        if filtered_logs:
            for log in reversed(filtered_logs):
                if log['status'] == cls.log_error:
                    error_count += 1
                    log_class = 'e-error'
                elif log['status'] == cls.log_warning:
                    warning_count += 1
                    log_class = 'e-warning'
                elif log['status'] == cls.log_success:
                    warning_count += 1
                    log_class = 'e-success'
                else:
                    log_class = ''
                formatted_messages.append(
                    f'<div class="log-entry" id="errors-logs"> <span class="log-entry-text" id={log["error_id"]}></span><span class="mdi mdi-alert e-{log["status"].lower()}"></span> <span>{log["date"]}</span> - <span class="{log_class}">{log["status"].upper()}</span> - <span> "{log["short_message"]}"</span></div>')
        #formatted_messages_str = "\n".join(formatted_messages)
        return {'error_messages': formatted_messages, 'error_count': error_count, 'warning_count': warning_count}

    @classmethod
    def load_logging_message_by_id(cls, error_id: str):
        """
        Load a logging message by its error_id from the error log file.

        Parameters:
        error_id (str): The ID of the error to load.

        Returns:
        dict: The logging message if found, None otherwise.
        """
        logging_dir = os.path.dirname(K2().path.data_path('conf/error_log/', True))
        filename = f"error.json"
        logging_file_dir = os.path.join(logging_dir, filename)

        if not os.path.exists(logging_file_dir):
            return None

        # Load existing logs
        with open(logging_file_dir, 'r') as file:
            logs = json.load(file)

        # Find log entry with matching error_id
        for log in logs:
            if log['error_id'] == error_id:
                return log

        # Return None if no log entry found with given error_id
        return None

    @classmethod
    def logging_message(cls, status, message, page_url=None, show_message=True):
        """
        Send a logging message to the client via Socket.IO.

        This method emits a message to the client with a given status and message content.
        The message is sent to a room identified by the current user's ID.


        Parameters:
        status (str): The status of the message (e.g.,  cls.log_error -  'ERROR', cls.log_warning - 'WARNING', cls.log_success -  'SUCCESS').
        message (str): The content of the message to be sent.
        page_url(str): url of page with Errors.
        show_message (bool): True  - save log and show windows,  False - only save_log

        Example:
        K2.logging_message(cls.log_error, 'Operation completed with error.')
        K2.logging_message(cls.log_success, 'Operation completed successfully.', 'users')
        K2.logging_message(cls.log_success, 'Operation completed successfully.', 'users', page_url='/kadm/users', show_message=False)

        """
        if page_url:
            name = cls.namemenu(page_url)
        status = status.upper()
        error_id = K2.generate_id()
        cls.save_logging_message(error_id, status, message, page_url)
        if status == cls.log_error:
            logging.error(message)
        elif status == cls.log_warning:
            logging.warning(message)
        elif status == cls.log_success:
            logging.info(message)
        from  app import socketio
        #

        log_entry = cls.log_entry(error_id, status, message, page_url)
        if current_user.get_id() in K2.connected_clients:
            sid = K2.connected_clients[current_user.get_id()]
            #socketio.emit(f'update_errors_count_result_{name}', message, room=sid)
            socketio.emit(f'update_errors_message_result_{name}', log_entry, room=sid)
        else:
            #socketio.emit(f'update_errors_count_result_{name}', message, room=current_user.get_id())
            socketio.emit(f'update_errors_message_result_{name}', log_entry, room=current_user.get_id())
        if show_message:
            if current_user.get_id() in K2.connected_clients:
                sid = K2.connected_clients[current_user.get_id()]
                socketio.emit(f'show_logging_message_{name}', {'status': status, 'message': message}, room=sid)
            else:
                socketio.emit(f'show_logging_message_{name}', {'status': status, 'message': message}, room=current_user.get_id())


    # @staticmethod
    def component_list():
        '''Пошук встановлених компонент'''
        with open(f"components/components.yml", 'r') as file:
            components_data = yaml.safe_load(file)
        if components_data is None or 'components' not in components_data:
            components_d = []
        else:
            components_d = components_data['components']
        installed_components = [{component['id']: component['name']} for component in components_d if
                                component.get('installed', False)]
        return installed_components

    components_install = component_list()  # список встановлених компонент

    # Отримуємо значення з path_objs для розширення поточного класу
    # @staticmethod
    def search_class_prop(path_objs):
        ''''''
        code_str = None
        for path_obj in path_objs:
            if os.path.exists(path_obj):
                with open(path_obj, 'r') as file:
                    code_str = file.read()
                break  # Виходимо з циклу, якщо знайдено файл
        else:
            sys.exit()  # Завершуємо процес, якщо жоден файл не знайдено
        return code_str

    code_str = K2Obj.search_class_prop(path_objs)


    try:
        exec(code_str)
    except Exception as e:
        print(f"Error in K2cfg {e}")
        sys.exit()
    domain_loc = f"http://127.0.0.1:{port}"

# print(K2.__dict__)
# print(K2.path_objs)

