import logging
import os
from datetime import datetime
import uuid
import re
import json

from flask_login import LoginManager, current_user
from flask import request, send_from_directory
from sqlalchemy.orm import sessionmaker

from k2.k2obj import K2Obj


class K2Settings(K2Obj):
    '''клас збереження та пошуку налаштувань  для користувача, складу, контрагента, проекту
    '''
    name = 'k2settings'
    def __init__(self, *args, **kwargs):
        super().__init__()

    #     self._data_path = None
    #
    # @property
    # def data_path(self):
    #     """Getter"""
    #     return self._data_path
    #
    # @data_path.setter
    # def name_yml(self, name: str):
    #     """Setter method"""
    #     self._data_path = name

    @classmethod
    def __generate_id(cls):
        '''Генерація ID'''
        generated_id = uuid.uuid4().hex[:32]
        return ''.join([generated_id[i:i + 4] for i in range(0, len(generated_id), 4)])

    def __current_proj(self, current_project_id):
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2Proj
            current_project = K2Proj.query.filter_by(projid=current_project_id).first()
        return current_project


    def __current_usr(self):
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2Proj, K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
        return user

    def __current_usr_role(self, user_id):
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2Proj, K2users
            user = K2users.query.get(user_id)
            roleid = user.roleid
        return roleid


    def __current_countreparts(self):
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2Proj, K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                current_counterpart_id = user.counterpart_id
            else:
                current_counterpart_id = None
        return current_counterpart_id


    def __current_stor(self):
        from app import app
        if hasattr(app, 'login_manager') and isinstance(app.login_manager,
                                                        LoginManager) and current_user.is_authenticated:
            from components.k2site.k2site.models import K2Proj, K2users
            user_id = current_user.get_id()
            user = K2users.query.get(user_id)
            if user is not None:
                current_storage_id = user.storage_id
            else:
                current_storage_id = None
        return current_storage_id

    def save_settings(self, settings, user_id, object, device=None, projid=None, counterpart_id=None, storage_id=None):
        '''
        Method to save settings for a user, storage, counterpart, or project.

        Args:
            settings (dict): A dictionary containing the settings to be saved. This will be converted to a JSON string.
            user_id (str): The ID of the user for whom the settings are being saved.
            object (str): The object for which the settings are specific (name of grid or form or other object).
            device (str, optional): The device for which the settings are specific, if any. Defaults to None.
            projid (str, optional): The project ID to which these settings apply, if any. Defaults to None.
            counterpart_id (str, optional): The counterpart ID to which these settings apply, if any. Defaults to None.
            storage_id (str, optional): The storage ID to which these settings apply, if any. Defaults to None.

        Examples:
            K2().settings.save_settings(settings=json_dict, user_id=current_user.get_id(), object='users')
            K2().settings.save_settings(settings=json_dict, user_id=current_user.get_id(), object='k2counterparts', projid=K2().get_user_project_id())

        Returns:
            None
        '''
        Session = sessionmaker(bind=self.db.engine)
        session = Session()

        try:
            from k2.k2cfg import K2
            from components.k2site.k2site.models import K2settings
            roleid = self.__current_usr_role(current_user.get_id())

            # Check if settings already exist
            existing_setting = session.query(K2settings).filter_by(
                user_id=user_id,
                object=object,
                projid=projid,
                counterpart_id=counterpart_id,
                storage_id=storage_id
            ).first()

            if existing_setting:
                # Update existing settings
                existing_setting.settings = json.dumps(settings)
                existing_setting.device = device
                existing_setting.updateuser = current_user.get_id()
                existing_setting.updatedate = datetime.now()
            else:
                # Create new settings
                new_setting = K2settings(
                    settingsid=self.__generate_id(),
                    roleid=roleid,
                    user_id=user_id,
                    device=device,
                    settings=json.dumps(settings),  # Convert the settings dictionary to a JSON string
                    object=object,
                    projid=projid,
                    counterpart_id=counterpart_id,
                    storage_id=storage_id,
                    createuser=current_user.get_id(),
                    createdate=datetime.now(),
                    updateuser=current_user.get_id(),
                    updatedate=datetime.now(),
                    active=1
                )
                session.add(new_setting)

            session.commit()
        except Exception as e:
            logging.error(f"Error in method save_settings {e}")
        finally:
            session.close()

    def get_settings(self, user_id, object):
        '''
        Method to get settings for a user according to the hierarchy of user, storage, counterpart, and project.

        Args:
            user_id (str): The ID of the user for whom the settings are being retrieved.
            object (str): The object for which the settings are specific (name of grid or form or other object).

        Returns:
            dict or None: The settings dictionary if found, otherwise None.
        '''
        Session = sessionmaker(bind=self.db.engine)
        session = Session()

        try:
            from components.k2site.k2site.models import K2settings, K2users
            user = session.query(K2users).get(user_id)
            if not user:
                return None

            # Check for user-specific settings
            user_setting = session.query(K2settings).filter_by(user_id=user_id, object=object).first()
            if user_setting:
                return json.loads(user_setting.settings)

            # Check for storage-specific settings
            storage_setting = session.query(K2settings).filter_by(storage_id=user.storage_id, object=object).first()
            if storage_setting:
                return json.loads(storage_setting.settings)

            # Check for counterpart-specific settings
            counterpart_setting = session.query(K2settings).filter_by(counterpart_id=user.counterpart_id,
                                                                      object=object).first()
            if counterpart_setting:
                return json.loads(counterpart_setting.settings)

            # Check for project-specific settings
            proj_setting = session.query(K2settings).filter_by(projid=user.projid, object=object).first()
            if proj_setting:
                return json.loads(proj_setting.settings)

            return None
        except Exception as e:
            logging.error(f"Error in method get_settings {e}")
            return None
        finally:
            session.close()
